#include "common.h"

void syserr(const char *fmt) {
	fprintf(stderr,"ERROR %s (%d; %s)\n", fmt, errno, strerror(errno));
	exit(1);
}

void err(char *msg) {
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

void do_run(int fd, int cfd, uint32_t addr, uint32_t size, int nxfd, int *xfd) {
	struct fdoomdev_ioctl_run run = {cfd, addr, size, nxfd};
	int i;
	for (i = 0; i < nxfd; i++)
		run.buffer_fd[i] = xfd[i];
	if (ioctl(fd, FDOOMDEV_IOCTL_RUN, &run))
		syserr("run");
}

void do_run_with_err(int fd, int cfd, uint32_t addr, uint32_t size, int nxfd, int *xfd) {
	struct fdoomdev_ioctl_run run = {cfd, addr, size, nxfd};
	int i;
	for (i = 0; i < nxfd; i++)
		run.buffer_fd[i] = xfd[i];
	if (ioctl(fd, FDOOMDEV_IOCTL_RUN, &run) != -1)
		err("run should fail");
}

void do_wait(int fd, uint32_t cnt) {
	struct fdoomdev_ioctl_wait wait = {cnt};
	if (ioctl(fd, FDOOMDEV_IOCTL_WAIT, &wait))
		syserr("wait");
}

void do_wait_for_err(int fd, uint32_t cnt) {
	struct fdoomdev_ioctl_wait wait = {cnt};
	if (ioctl(fd, FDOOMDEV_IOCTL_WAIT, &wait) != -1)
		err("wait should fail");
}


void do_run_and_wait_with_err(int fd, int cfd, uint32_t addr, uint32_t size, int nxfd, int *xfd) {
	struct fdoomdev_ioctl_run run = {cfd, addr, size, nxfd};
	int i;
	for (i = 0; i < nxfd; i++)
		run.buffer_fd[i] = xfd[i];
	ioctl(fd, FDOOMDEV_IOCTL_RUN, &run); // ignore
	do_wait_for_err(fd, 0);
}

int do_open(char *path) {
	int fd;
	if ((fd = open(path, O_RDWR)) < 0)
		syserr("open");
	return fd;
}

void do_close(int fd) {
	if (close(fd) < 0)
		syserr("close");
}

void do_munmap(void *addr, size_t len) {
	if (munmap(addr, len) < 0)
		syserr("munmap");
}

int do_open0() {
	return do_open("/dev/fdoom0");
}

int do_open1() {
	return do_open("/dev/fdoom1");
}

int do_create_buf(int fd, int size, int pitch) {
	struct fdoomdev_ioctl_create_buffer cb = { size, pitch };
	int bfd;
	if ((bfd = ioctl(fd, FDOOMDEV_IOCTL_CREATE_BUFFER, &cb)) < 0)
		syserr("create_buffer");
	return bfd;
}

void do_create_buf_with_err(int fd, int size, int pitch) {
	struct fdoomdev_ioctl_create_buffer cb = { size, pitch };
	if (ioctl(fd, FDOOMDEV_IOCTL_CREATE_BUFFER, &cb) != -1)
		err("create buffer should fail");
}

void do_resize_buf(int fd, int size) {
	struct fdoomdev_ioctl_buffer_resize rb = { size };
	if (ioctl(fd, FDOOMDEV_IOCTL_BUFFER_RESIZE, &rb))
		syserr("resize_buffer");
}

void do_resize_buf_with_err(int fd, int size) {
	struct fdoomdev_ioctl_buffer_resize rb = { size };
	if (ioctl(fd, FDOOMDEV_IOCTL_BUFFER_RESIZE, &rb) != -1)
		syserr("resize_buffer should fail");
}
